# Functional example of weak form implementation.

import netgen.gui
from ngsolve import *
from netgen.csg import *
import numpy as np

# Generate mesh:
mesh_size = 1/2
geo = CSGeometry()

cube = OrthoBrick( Pnt(-1,-1,-1), Pnt(1,1,1) ).bc("outer") # cube
geo.Add (cube)
mesh = Mesh(geo.GenerateMesh(maxh=mesh_size)) 

# Generate function space:
ord = 1
fesm = HCurl(mesh, order=ord, complex=True, dirichlet="outer")
fes = FESpace([fesm,fesm])

# Define test and trial functions:
trial_A, trial_B = fes.TrialFunction()
test_A, test_B = fes.TestFunction()

# Define constants:
piee = np.pi

c = CoefficientFunction(1)
d = CoefficientFunction(1)
Q_1 = CoefficientFunction((1, 0, 0)) 
Q_2 = CoefficientFunction((0, 0, 1)) 
F_1 = CoefficientFunction((0, 1, 0))
F_2 = CoefficientFunction((1, 0, 0)) 

# Dirichlet boundary condition:
bc1 = CoefficientFunction((0, 0, 0)) 
bc2 = CoefficientFunction((0, 0, 0)) 


# Variational problem:

a = BilinearForm(fes, symmetric=True, eliminate_internal=True)
a += SymbolicBFI(1j*InnerProduct(c*trial_A, test_B)+ 1j*(InnerProduct(d*trial_A, test_A))) 
a += SymbolicBFI(-InnerProduct(trial_B, curl(test_B)) + InnerProduct(trial_A, curl(test_A))) 
        
f = LinearForm(fes)
f += SymbolicLFI(InnerProduct(-Q_2,test_B) + InnerProduct(-Q_1, test_A ))
f += SymbolicLFI(InnerProduct(F_2, test_B.Trace()), BND)
f += SymbolicLFI(InnerProduct(-F_1, test_A.Trace()), BND)

# Define preconditioner. Solution depends a lot on this.
    # Options ["local","direct","multigrid","bddc"]
precond = "local"
c = Preconditioner(a, type=precond)#, inverse="sparsecholesky")

# Generate solution space; Set boundary conditions:
u = GridFunction(fes)
u1, u2 = u.components
u1.Set(bc1, BND)
u2.Set(bc2, BND)

# Solve problem:
with TaskManager():
    a.Assemble()
    f.Assemble()
    solver = CGSolver(mat=a.mat, pre=c.mat, complex=True)
    u.vec.data = solver * f.vec            
    # Simplest way (no other solvers): 
    # u.vec.data = a.mat.Inverse(fes.FreeDofs()) * f.vec 
    
# Plot:  
Draw (u.components[0], mesh, name="numA")
Draw (u.components[1], mesh, name="numB") 

# Print entries of solution vectors for sanity check:
print(u.components[0].vec) 
print(u.components[1].vec) 
